import argparse
import cv2
import glob
import numpy as np
from collections import OrderedDict
import os
import torch
import requests

from models.network_swinir import SwinIR as net
from utils import util_calculate_psnr_ssim as util
from imresize import imresize
import time
def main():
    start_time = time.time()

    workspace = os.path.dirname(os.path.realpath(__file__))
    #model_ckpt_path = os.path.join(workspace, "demo/hs_triplet_upscaling.model")
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--task', type=str, default='real_sr', help='real_sr, classical_sr')
    parser.add_argument('--model_path', type=str, default='')
    parser.add_argument('--input', type=str, default='input_0.png', help='input test image')
    parser.add_argument('--tile', type=int, default=None, help='Tile size, None for no tile during testing (testing as a whole)')
    parser.add_argument('--tile_overlap', type=int, default=32, help='Overlapping of different tiles')
    
    parser.add_argument('--degrade', type=str, default="False")

    
    parser.add_argument('--downsample', type=str, default="False")
    #parser.add_argument('--add_noise', type=str, default="False")
    parser.add_argument('--noise', type=int, default=0, help='noise level: 15, 25, 50')
    
    #parser.add_argument('--jpeg_compression', type=str, default="False")
    parser.add_argument('--jpeg', type=int, default=40, help='scale factor: 10, 20, 30, 40')

    args = parser.parse_args()

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    # set up model
    args.degrade = (args.degrade.lower() == 'true')
    args.downsample = (args.downsample.lower() == 'true')
    #args.add_noise = (args.add_noise.lower() == 'true')

    #args.jpeg_compression = (args.jpeg_compression.lower() == 'true')
    
    
    model_name = get_model_name(args)
    
    args = inquire(args, model_name, workspace)

    
    model = define_model(args)
    model.eval()
    model = model.to(device)


    # setup folder and path
    border, window_size = setup(args)
    test_results = OrderedDict()
    test_results['psnr'] = []
    test_results['ssim'] = []
    test_results['psnr_y'] = []
    test_results['ssim_y'] = []
    test_results['psnr_b'] = []
    psnr, ssim, psnr_y, ssim_y, psnr_b = 0, 0, 0, 0, 0

    # read image
    imgname, img_lq, img_gt = get_image_pair(args)  # image to HWC-BGR, float32
    #if args.downsample:
    kron_ker = np.ones((4,4,1))

    img_lq = np.clip(img_lq, 0, 1)
    cv2.imwrite('degraded.png', (img_lq * 255.).round().astype(np.uint8)) 
    cv2.imwrite('degraded_zoom.png', (np.kron(img_lq, kron_ker) * 255.).round().astype(np.uint8)) #better visual comparison

    img_lq = np.transpose(img_lq if img_lq.shape[2] == 1 else img_lq[:, :, [2, 1, 0]], (2, 0, 1))  # HCW-BGR to CHW-RGB
    img_lq = torch.from_numpy(img_lq).float().unsqueeze(0).to(device)  # CHW-RGB to NCHW-RGB

    # inference
    with torch.no_grad():
        # pad input image to be a multiple of window_size
        _, _, h_old, w_old = img_lq.size()
        h_pad = (h_old // window_size + 1) * window_size - h_old
        w_pad = (w_old // window_size + 1) * window_size - w_old
        img_lq = torch.cat([img_lq, torch.flip(img_lq, [2])], 2)[:, :, :h_old + h_pad, :]
        img_lq = torch.cat([img_lq, torch.flip(img_lq, [3])], 3)[:, :, :, :w_old + w_pad]
        output = test(img_lq, model, args, window_size)
        output = output[..., :h_old * 4, :w_old * 4]

    # save image
    output = output.data.squeeze().float().cpu().clamp_(0, 1).numpy()
    if output.ndim == 3:
        output = np.transpose(output[[2, 1, 0], :, :], (1, 2, 0))  # CHW-RGB to HCW-BGR
    output = (output * 255.0).round().astype(np.uint8)  # float32 to uint8
    cv2.imwrite('output.png', output)



    # evaluate psnr/ssim/psnr_b
    if img_gt is not None:
        if img_gt.shape != output.shape:
            print("To compute the PSNR and SSIM, make sure that the shape of the input image is multiple of 4")
        else:
            img_gt = (img_gt * 255.0).round().astype(np.uint8)  # float32 to uint8
            #print(img_gt.shape, output.shape)
            img_gt = img_gt[:h_old * 4, :w_old * 4, ...]  # crop gt
            img_gt = np.squeeze(img_gt)
            #print(img_gt.shape)

            psnr = util.calculate_psnr(output, img_gt, crop_border=border)
            ssim = util.calculate_ssim(output, img_gt, crop_border=border)
            test_results['psnr'].append(psnr)
            test_results['ssim'].append(ssim)
            if img_gt.ndim == 3:  # RGB image
                psnr_y = util.calculate_psnr(output, img_gt, crop_border=border, test_y_channel=True)
                ssim_y = util.calculate_ssim(output, img_gt, crop_border=border, test_y_channel=True)
                test_results['psnr_y'].append(psnr_y)
                test_results['ssim_y'].append(ssim_y)
            print('PSNR: {:.2f} dB; SSIM: {:.4f}; PSNR_Y: {:.2f} dB; SSIM_Y: {:.4f} dB.'.format(psnr, ssim, psnr_y, ssim_y))
    else:
        #print('Testing {:20s}'.format(imgname))
        pass

    print("Execution time: {:.1f}s ".format(time.time() - start_time))


def define_model(args):
    # 001 classical image sr
    if args.task == 'classical_sr':
        model = net(upscale=4, in_chans=3, img_size=64, window_size=8,
                    img_range=1., depths=[6, 6, 6, 6, 6, 6], embed_dim=180, num_heads=[6, 6, 6, 6, 6, 6],
                    mlp_ratio=2, upsampler='pixelshuffle', resi_connection='1conv')
        param_key_g = 'params'

    else:
        # larger model size; use '3conv' to save parameters and memory; use ema for GAN training
        model = net(upscale=4, in_chans=3, img_size=64, window_size=8,
                    img_range=1., depths=[6, 6, 6, 6, 6, 6, 6, 6, 6], embed_dim=240,
                    num_heads=[8, 8, 8, 8, 8, 8, 8, 8, 8],
                    mlp_ratio=2, upsampler='nearest+conv', resi_connection='3conv')
        param_key_g = 'params_ema'
        
    pretrained_model = torch.load(os.path.join(args.model_path))
    model.load_state_dict(pretrained_model[param_key_g] if param_key_g in pretrained_model.keys() else pretrained_model, strict=True)
    
    return model


def setup(args):
    # 001 classical image sr/ 002 lightweight image sr
    if args.task == 'classical_sr':
        border = 4
        window_size = 8

    # 003 real-world image sr
    else:
        border = 0
        window_size = 8

    return border, window_size


def get_image_pair(args):
    (imgname, imgext) = os.path.splitext(os.path.basename(args.input))

    # 001 classical image sr/ 002 lightweight image sr (load lq-gt image pairs)
    if args.degrade:
        if args.downsample:
            img_gt = cv2.imread(args.input, cv2.IMREAD_COLOR).astype(np.float32) / 255.
            img_lq = np.clip(imresize(img_gt, 1./4), 0, 1)
        else:
            img_gt = None
            img_lq = cv2.imread(args.input, cv2.IMREAD_COLOR).astype(np.float32) / 255.
        if args.noise >0:
            np.random.seed(seed=0)
            img_lq = img_lq + np.random.normal(0, args.noise / 255., img_lq.shape)
        if args.jpeg < 100:
            result, encimg = cv2.imencode('.jpg', (np.clip(img_lq, 0, 1)*255).astype(np.uint8), [int(cv2.IMWRITE_JPEG_QUALITY), args.jpeg])
            img_lq = cv2.imdecode(encimg, 1) /255.

    else:
        img_gt = None
        img_lq = cv2.imread(args.input, cv2.IMREAD_COLOR).astype(np.float32) / 255.

    return imgname, img_lq, img_gt


def test(img_lq, model, args, window_size):
    if args.tile is None:
        # test the image as a whole
        output = model(img_lq)
    else:
        # test the image tile by tile
        b, c, h, w = img_lq.size()
        tile = min(args.tile, h, w)
        assert tile % window_size == 0, "tile size should be a multiple of window_size"
        tile_overlap = args.tile_overlap
        sf = 4

        stride = tile - tile_overlap
        h_idx_list = list(range(0, h-tile, stride)) + [h-tile]
        w_idx_list = list(range(0, w-tile, stride)) + [w-tile]
        E = torch.zeros(b, c, h*sf, w*sf).type_as(img_lq)
        W = torch.zeros_like(E)

        for h_idx in h_idx_list:
            for w_idx in w_idx_list:
                in_patch = img_lq[..., h_idx:h_idx+tile, w_idx:w_idx+tile]
                out_patch = model(in_patch)
                out_patch_mask = torch.ones_like(out_patch)

                E[..., h_idx*sf:(h_idx+tile)*sf, w_idx*sf:(w_idx+tile)*sf].add_(out_patch)
                W[..., h_idx*sf:(h_idx+tile)*sf, w_idx*sf:(w_idx+tile)*sf].add_(out_patch_mask)
        output = E.div_(W)

    return output

def inquire(args, model_name, workspace):
    args.model_path = os.path.join(workspace, "model_zoo/swinir", model_name)
    if os.path.exists(args.model_path):
        #print(f'loading model from {args.model_path}')
        pass
    else:
        os.makedirs(os.path.dirname(args.model_path), exist_ok=True)
        url = 'https://github.com/JingyunLiang/SwinIR/releases/download/v0.0/{}'.format(os.path.basename(args.model_path))
        r = requests.get(url, allow_redirects=True)
        print(f'downloading model {args.model_path}')
        open(args.model_path, 'wb').write(r.content)
        
    return args

def get_model_name(args):
    if args.task == 'classical_sr':
        model_name = "001_classicalSR_DF2K_s64w8_SwinIR-M_x4.pth"
    else:
        model_name = "003_realSR_BSRGAN_DFOWMFC_s64w8_SwinIR-L_x4_GAN.pth" #Real SR
    return model_name


if __name__ == '__main__':
    main()
